﻿using System;
using System.Collections.Generic;
using JetBrains.Annotations;
using Orchard.Caching;
using Orchard.ContentManagement;
using Orchard.Core.Navigation.Models;
using Orchard.Localization;
using Orchard.UI.Navigation;
using System.Linq;
using PS.Library.Collections;

namespace Szmyd.Orchard.Modules.Menu.Services
{
    [UsedImplicitly]
    public class HierarchicalMainMenuNavigationProvider : INavigationProvider
    {
        private readonly IContentManager _contentManager;
        private readonly ICacheManager _cacheManager;

        public HierarchicalMainMenuNavigationProvider(IContentManager contentManager, ICacheManager cacheManager)
        {
            _contentManager = contentManager;
            _cacheManager = cacheManager;
        }

        public string MenuName { get { return "main"; } }

        public void GetNavigation(NavigationBuilder builder)
        {
            // Preparing trie (prefix tree) for keeping menu hierarchy (if not in cache)
            var tree = GetTrie();// _cacheManager.Get("MenuHierarchyTrie", ctx => GetTrie());

            // Preparing the build action (if not in cache)
            var buildAction = GetBuildAction(tree);// _cacheManager.Get("MenuHierarchyBuildAction", ctx => GetBuildAction(tree));

            // Building menu
            builder.Add(buildAction);
        }

        protected internal Action<NavigationItemBuilder> GetBuildAction(Trie<MenuPart> tree)
        {
            // Anonymous function for recurrent action building
            // Unfortunately no tail-recurrent...:/
            Func<TrieNode<MenuPart>, Action<NavigationItemBuilder>> funcBuild = null;
            funcBuild = item =>
            {
                var nestedActions = item.Children.Select(c => funcBuild(c.Value));
                
                Action<NavigationItemBuilder> cAction = null;
                if(nestedActions.Any())
                    cAction = nib => nestedActions.Where(a => a != null).ToList().ForEach(a => a(nib));
                
                Action<NavigationItemBuilder> action = null;

                if (item.Values.Any())
                {

                    // Adds current item (some action) info into current builder level)
                    foreach (var i in item.Values)
                    {
                        var text = i.MenuText;
                        var position = i.MenuPosition;
                        if (i.Is<MenuItemPart>())
                        {
                            var url = i.As<MenuItemPart>().Url;
                            action = parentBuilder => parentBuilder.Add(new LocalizedString(text), position,
                                                                        nib2 => {
                                                                            nib2.Url(url); 
                                                                            if (cAction != null) cAction(nib2);
                                                                        });
                        }
                        else
                        {
                            var contentItem = i.ContentItem;
                            action = parentBuilder => parentBuilder.Add(new LocalizedString(text), position,
                                                                        nib2 => {
                                                                            nib2.Action(_contentManager.GetItemMetadata(contentItem).DisplayRouteValues);
                                                                            if (cAction != null) cAction(nib2);
                                                                        });
                        }
                    }
                }

                return action ?? cAction;
            };

            return tree.Execute(funcBuild, false);
        }

        protected internal Trie<MenuPart> GetTrie()
        {
            var menuParts = _contentManager
                    .Query<MenuPart, MenuPartRecord>()
                    .Where(x => x.OnMainMenu)
                    .List();

            // Building menu items prefix tree containing appropriate builder actions
            var t = new Trie<MenuPart>(new[] { '.', ',', ':', '/', '\\' });
            foreach (var part in menuParts.Where(part => part != null))
            {
                t[part.MenuPosition] = new[] { part };
            }
            return t;
        }
    }

    public static class EnumerableExtensions
    {
        public static IList<IGrouping<TKey, T>> Partition<T, TKey>(this IEnumerable<T> source, Func<T, TKey> selector)
        {
            return source.GroupBy(selector, e => e).OrderBy(e => e.Key).ToList();
        }
    }

    ///// <summary>
    ///// Internal class for creating menu tree.
    ///// </summary>
    //internal class TreeBuilder<T, TKey>
    //{

    //    public TreeBuilder()
    //    {
    //        Children = new List<TreeBuilder<T, TKey>>();
    //        Parent = null;
    //        Item = default(T);
    //    }

    //    /// <summary>
    //    /// Builds tree from  given collection.
    //    /// </summary>
    //    /// <param name="items">Item collection from which to build the tree.</param>
    //    /// <param name="keySelector">Key selector function.</param>
    //    /// <param name="parentSelector">Function defining parent-child relation.</param>
    //    public TreeBuilder(IEnumerable<T> items, Func<T, TKey> keySelector, Func<T, bool> parentSelector)
    //    {
    //        int i = 0;
    //        foreach (var item in items.Select(parentSelector))
    //        {
    //            Children.Add(new TreeBuilder<T, TKey>());
    //            i++;
    //        }
    //    }

    //    public TKey Id { get; set; }
    //    /// <summary>
    //    /// Corresponding item
    //    /// </summary>
    //    public T Item { get; set; }

    //    /// <summary>
    //    /// Parent node (null for root)
    //    /// </summary>
    //    public TreeBuilder Parent { get; set; }

    //    /// <summary>
    //    /// Children nodes. Root always has children.
    //    /// </summary>
    //    public IList<TreeBuilder> Children { get; set; }

    //    /// <summary>
    //    /// Current item build action for displaying item.
    //    /// </summary>
    //    public Action<NavigationItemBuilder> Action { get; set; }

    //    /// <summary>
    //    /// Recurrent buildup of menu item hierarchy.
    //    /// </summary>
    //    /// <param name="parentBuilder">Builder on which this and child nodes' actions will execute.</param>
    //    /// <returns></returns>
    //    public Action<NavigationItemBuilder> BuildAction(NavigationBuilder parentBuilder)
    //    {
    //        Action<NavigationItemBuilder> a = null;
    //        if (Children.Any())
    //        {

    //            foreach (var c in Children)
    //            {
    //                var c1 = c;
    //                a = nib => nib.Add(c1.BuildAction(nib));
    //            }
    //        }

    //        if (a != null && Action != null)
    //        {
    //            parentBuilder.Add(Action);
    //        }

    //        return a ?? Action;

    //    }
    //}


}